@@ -17,7 +17,6 @@ from datashape import coretypes as ct
 from collections import namedtuple, Iterator
 from contextlib import contextmanager
 from .ssh import SSH
-from .spark import SparkDataFrame, SchemaRDD
 from .sql import metadata_of_engine, sa
 from ..utils import tmpfile, sample, ignoring, raises
 from ..temp import Temp
@@ -157,9 +156,11 @@ def resource_hive_table(uri, stored_as='TEXTFILE', external=True, dshape=None, *
                         autoload_with=engine)
 
     # Enough information to make an internal table
-    if dshape and not external:
+    if dshape and (not external or external and kwargs.get('path')):
+        table_type = 'EXTERNAL' if external else ''
         statement = create_hive_statement(table, dshape,
-                db_name=engine.url.database, stored_as=stored_as, **kwargs)
+                db_name=engine.url.database, stored_as=stored_as,
+                table_type=table_type, **kwargs)
         with engine.connect() as conn:
             conn.execute(statement)
         return sa.Table(table, metadata, autoload=True,
@@ -451,7 +452,7 @@ def convert_hdfs_file_to_temp_local(source, **kwargs):
 @append.register(HDFS(TextFile), TextFile)
 @append.register(HDFS(JSONLines), JSONLines)
 @append.register(HDFS(JSON), JSON)
-@append.register(HDFS(CSV), CSV)
+@append.register(HDFS(CSV), (CSV, Temp(CSV)))
 def append_local_file_to_hdfs(target, source, blocksize=100000, **kwargs):
     if raises(FileNotFound,
               lambda: target.hdfs.list_dir(target.path.lstrip('/'))):
@@ -535,8 +536,6 @@ def resource_hdfs(uri, **kwargs):
 
 
 @append.register(HDFS(TextFile), (Iterator, object))
-@append.register(HDFS(JSONLines), (Iterator, object, SparkDataFrame,
-                                   SchemaRDD))
 @append.register(HDFS(JSON), (list, object))
 @append.register(HDFS(CSV), (chunks(pd.DataFrame), pd.DataFrame, object))
 def append_object_to_hdfs(target, source, **kwargs):
