@@ -1,10 +1,12 @@
 from __future__ import absolute_import, division, print_function
 
-from operator import attrgetter
 import os
 import re
 import subprocess
+import sys
+import decimal
 
+from operator import attrgetter
 from itertools import chain
 from collections import Iterator
 from datetime import datetime, date
@@ -22,7 +24,8 @@ from multipledispatch import MDNotImplementedError
 import datashape
 from datashape import DataShape, Record, Option, var, dshape, Map
 from datashape.predicates import isdimension, isrecord, isscalar
-from datashape import discover
+from datashape import discover, datetime_, date_, float64, int64, int_, string
+from datashape import float32
 from datashape.dispatch import dispatch
 
 from toolz import (partition_all, keyfilter, memoize, valfilter, identity,
@@ -37,65 +40,55 @@ from ..resource import resource
 from ..chunks import Chunks
 from .csv import CSV
 
-base = (int, float, datetime, date, bool, str)
+base = int, float, datetime, date, bool, str, decimal.Decimal
 
 
 # http://docs.sqlalchemy.org/en/latest/core/types.html
 
 types = {
-    'int64': sa.types.BigInteger,
-    'int32': sa.types.Integer,
-    'int': sa.types.Integer,
-    'int16': sa.types.SmallInteger,
-    'float32': sa.types.Float(precision=24),  # sqlalchemy uses mantissa
-    'float64': sa.types.Float(precision=53),  # for precision
-    'float': sa.types.Float(precision=53),
-    'real': sa.types.Float(precision=53),
-    'string': sa.types.Text,
-    'date': sa.types.Date,
-    'time': sa.types.Time,
-    'datetime': sa.types.DateTime,
-    'bool': sa.types.Boolean,
-    "timedelta[unit='D']": sa.types.Interval(second_precision=0,
-                                             day_precision=9),
-    "timedelta[unit='h']": sa.types.Interval(second_precision=0,
-                                             day_precision=0),
-    "timedelta[unit='m']": sa.types.Interval(second_precision=0,
-                                             day_precision=0),
-    "timedelta[unit='s']": sa.types.Interval(second_precision=0,
-                                             day_precision=0),
-    "timedelta[unit='ms']": sa.types.Interval(second_precision=3,
-                                              day_precision=0),
-    "timedelta[unit='us']": sa.types.Interval(second_precision=6,
-                                              day_precision=0),
-    "timedelta[unit='ns']": sa.types.Interval(second_precision=9,
-                                              day_precision=0),
+    'int64': sa.BigInteger,
+    'int32': sa.Integer,
+    'int': sa.Integer,
+    'int16': sa.SmallInteger,
+    'float32': sa.REAL,
+    'float64': sa.FLOAT,
+    'float': sa.FLOAT,
+    'real': sa.FLOAT,
+    'string': sa.Text,
+    'date': sa.Date,
+    'time': sa.Time,
+    'datetime': sa.DateTime,
+    'bool': sa.Boolean,
+    "timedelta[unit='D']": sa.Interval(second_precision=0, day_precision=9),
+    "timedelta[unit='h']": sa.Interval(second_precision=0, day_precision=0),
+    "timedelta[unit='m']": sa.Interval(second_precision=0, day_precision=0),
+    "timedelta[unit='s']": sa.Interval(second_precision=0, day_precision=0),
+    "timedelta[unit='ms']": sa.Interval(second_precision=3, day_precision=0),
+    "timedelta[unit='us']": sa.Interval(second_precision=6, day_precision=0),
+    "timedelta[unit='ns']": sa.Interval(second_precision=9, day_precision=0),
     # ??: sa.types.LargeBinary,
-    # Decimal: sa.types.Numeric,
-    # ??: sa.types.PickleType,
-    # unicode: sa.types.Unicode,
-    # unicode: sa.types.UnicodeText,
-    # str: sa.types.Text,  # ??
 }
 
 revtypes = dict(map(reversed, types.items()))
 
 revtypes.update({
-    sa.types.DATETIME: 'datetime',
-    sa.types.TIMESTAMP: 'datetime',
-    sa.types.FLOAT: 'float64',
-    sa.types.DATE: 'date',
-    sa.types.BIGINT: 'int64',
-    sa.types.INTEGER: 'int',
-    sa.types.NUMERIC: 'float64',  # TODO: extend datashape to decimal
-    sa.types.BIGINT: 'int64',
-    sa.types.NullType: 'string',
-    sa.types.Float: 'float64',
+    sa.DATETIME: datetime_,
+    sa.TIMESTAMP: datetime_,
+    sa.FLOAT: float64,
+    sa.DATE: date_,
+    sa.BIGINT: int64,
+    sa.INTEGER: int_,
+    sa.BIGINT: int64,
+    sa.types.NullType: string,
+    sa.REAL: float32,
+    sa.Float: float64,
+    sa.Float(precision=24): float32,
+    sa.Float(precision=53): float64,
 })
 
 # interval types are special cased in discover_typeengine so remove them from
 # revtypes
-revtypes = valfilter(lambda x: not isinstance(x, sa.types.Interval), revtypes)
+revtypes = valfilter(lambda x: not isinstance(x, sa.Interval), revtypes)
 
 
 units_of_power = {
@@ -110,7 +103,16 @@ sa.dialects.registry.load('oracle')
 sa.dialects.registry.load('postgresql')
 
 
-def batch(sel, chunksize=10000):
+def getbind(t, bind):
+    if bind is None:
+        return t.bind
+
+    if isinstance(bind, sa.engine.base.Engine):
+        return bind
+    return sa.create_engine(bind)
+
+
+def batch(sel, chunksize=10000, bind=None):
     """Execute `sel`, streaming row at a time and fetching from the database in
     batches of size `chunksize`.
 
@@ -122,7 +124,7 @@ def batch(sel, chunksize=10000):
         Number of rows to fetch from the database
     """
     def rowterator(sel, chunksize=chunksize):
-        with sel.bind.connect() as conn:
+        with getbind(sel, bind).connect() as conn:
             result = conn.execute(sel)
             yield result.keys()
 
@@ -138,18 +140,17 @@ def batch(sel, chunksize=10000):
 
 @discover.register(sa.dialects.postgresql.base.INTERVAL)
 def discover_postgresql_interval(t):
-    return discover(sa.types.Interval(day_precision=0,
-                                      second_precision=t.precision))
+    return discover(sa.Interval(day_precision=0, second_precision=t.precision))
 
 
 @discover.register(sa.dialects.oracle.base.INTERVAL)
 def discover_oracle_interval(t):
-    return discover(t.adapt(sa.types.Interval))
+    return discover(t.adapt(sa.Interval))
 
 
 @discover.register(sa.sql.type_api.TypeEngine)
 def discover_typeengine(typ):
-    if isinstance(typ, sa.types.Interval):
+    if isinstance(typ, sa.Interval):
         if typ.second_precision is None and typ.day_precision is None:
             return datashape.TimeDelta(unit='us')
         elif typ.second_precision == 0 and typ.day_precision == 0:
@@ -168,6 +169,8 @@ def discover_typeengine(typ):
         return dshape(revtypes[typ])[0]
     if type(typ) in revtypes:
         return revtypes[type(typ)]
+    if isinstance(typ, (sa.NUMERIC, sa.DECIMAL)):
+        return datashape.Decimal(precision=typ.precision, scale=typ.scale)
     if isinstance(typ, (sa.String, sa.Unicode)):
         return datashape.String(typ.length, typ.collation)
     else:
@@ -372,45 +375,45 @@ def dshape_to_alchemy(dshape, primary_key=frozenset()):
     if isinstance(dshape, datashape.String):
         fixlen = dshape[0].fixlen
         if fixlen is None:
-            return sa.types.Text
-        string_types = dict(U=sa.types.Unicode, A=sa.types.String)
+            return sa.TEXT
+        string_types = dict(U=sa.Unicode, A=sa.String)
         assert dshape.encoding is not None
         return string_types[dshape.encoding[0]](length=fixlen)
     if isinstance(dshape, datashape.DateTime):
-        if dshape.tz:
-            return sa.types.DateTime(timezone=True)
-        else:
-            return sa.types.DateTime(timezone=False)
+        return sa.DATETIME(timezone=dshape.tz is not None)
+    if isinstance(dshape, datashape.Decimal):
+        return sa.NUMERIC(dshape.precision, dshape.scale)
     raise NotImplementedError("No SQLAlchemy dtype match for datashape: %s"
                               % dshape)
 
 
 @convert.register(Iterator, sa.Table, cost=300.0)
-def sql_to_iterator(t, **kwargs):
-    _, rows = batch(sa.select([t]))
+def sql_to_iterator(t, bind=None, **kwargs):
+    _, rows = batch(sa.select([t]), bind=bind)
     return map(tuple, rows)
 
 
 @convert.register(Iterator, sa.sql.Select, cost=300.0)
-def select_to_iterator(sel, dshape=None, **kwargs):
+def select_to_iterator(sel, dshape=None, bind=None, **kwargs):
     func = pluck(0) if dshape and isscalar(dshape.measure) else map(tuple)
-    _, rows = batch(sel)
+    _, rows = batch(sel, bind=bind)
     return func(rows)
 
 
 @convert.register(base, sa.sql.Select, cost=300.0)
-def select_to_base(sel, dshape=None, **kwargs):
-    assert not dshape or isscalar(dshape), \
-        'dshape should be None or scalar, got %s' % dshape
-    with sel.bind.connect() as conn:
+def select_to_base(sel, dshape=None, bind=None, **kwargs):
+    if dshape is not None and not isscalar(dshape):
+        raise ValueError('dshape should be None or scalar, got %s' % dshape)
+    with getbind(sel, bind).connect() as conn:
         return conn.execute(sel).scalar()
 
 
 @append.register(sa.Table, Iterator)
-def append_iterator_to_table(t, rows, dshape=None, **kwargs):
+def append_iterator_to_table(t, rows, dshape=None, bind=None, **kwargs):
     assert not isinstance(t, type)
-    if not t.exists():
-        raise ValueError('table %r does not exist' % t.name)
+    engine = getbind(t, bind)
+    if not t.exists(bind=engine):
+        t.create(bind=engine)
     rows = iter(rows)
 
     # We see if the sequence is of tuples or dicts
@@ -421,6 +424,7 @@ def append_iterator_to_table(t, rows, dshape=None, **kwargs):
         return
     rows = chain([row], rows)
     if isinstance(row, (tuple, list)):
+        dshape = dshape and datashape.dshape(dshape)
         if dshape and isinstance(dshape.measure, datashape.Record):
             names = dshape.measure.names
             if set(names) != set(discover(t).measure.names):
@@ -433,7 +437,6 @@ def append_iterator_to_table(t, rows, dshape=None, **kwargs):
             names = discover(t).measure.names
         rows = (dict(zip(names, row)) for row in rows)
 
-    engine = t.bind
     with engine.connect() as conn:
         for chunk in partition_all(1000, rows):  # TODO: 1000 is hardcoded
             conn.execute(t.insert(), chunk)
@@ -455,29 +458,29 @@ def append_anything_to_sql_Table(t, o, **kwargs):
 
 @append.register(sa.Table, sa.Table)
 def append_table_to_sql_Table(t, o, **kwargs):
-    # This condition is an ugly kludge and should be removed once
-    # https://github.com/dropbox/PyHive/issues/15 is resolved
-    if t.bind.name == o.bind.name == 'hive':
-        with t.bind.connect() as conn:
-            conn.execute('INSERT INTO TABLE %s SELECT * FROM %s' %
-                         (t.name, o.name))
-        return t
-
     s = sa.select([o])
     return append(t, s, **kwargs)
 
 
 @append.register(sa.Table, sa.sql.Select)
-def append_select_statement_to_sql_Table(t, o, **kwargs):
-    if not o.bind == t.bind:
-        return append(t, convert(Iterator, o, **kwargs), **kwargs)
+def append_select_statement_to_sql_Table(t, o, bind=None, **kwargs):
+    t_bind = getbind(t, bind)
+    o_bind = getbind(o, bind)
+    if t_bind != o_bind:
+        return append(
+            t,
+            convert(Iterator, o, bind=bind, **kwargs),
+            bind=bind,
+            **kwargs
+        )
+    bind = t_bind
 
-    assert o.bind.has_table(t.name, t.schema), \
+    assert bind.has_table(t.name, t.schema), \
         'tables must come from the same database'
 
     query = t.insert().from_select(o.columns.keys(), o)
 
-    with o.bind.connect() as conn:
+    with bind.connect() as conn:
         conn.execute(query)
     return t
 
@@ -490,22 +493,17 @@ def should_create_schema(ddl, target, bind, tables=None, state=None,
 def attach_schema(obj, schema):
     if schema is not None:
         ddl = CreateSchema(schema, quote=True)
-        event.listen(obj,
-                     'before_create',
-                     ddl.execute_if(callable_=should_create_schema,
-                                    dialect='postgresql'))
+        event.listen(
+            obj,
+            'before_create',
+            ddl.execute_if(
+                callable_=should_create_schema,
+                dialect='postgresql'
+            )
+        )
     return obj
 
 
-def fullname(table, compiler):
-    preparer = compiler.dialect.identifier_preparer
-    fullname = preparer.quote_identifier(table.name)
-    schema = table.schema
-    if schema is not None:
-        fullname = '%s.%s' % (preparer.quote_schema(schema), fullname)
-    return fullname
-
-
 @resource.register(r'(.*sql.*|oracle|redshift)(\+\w+)?://.+')
 def resource_sql(uri, *args, **kwargs):
     engine = create_engine(uri, **filter_kwargs(sa.create_engine, kwargs))
@@ -590,15 +588,16 @@ ooc_types.add(sa.Table)
 
 
 @dispatch(sa.Table)
-def drop(table):
-    table.drop(table.bind, checkfirst=True)
-    if table.exists():
+def drop(table, bind=None):
+    bind = getbind(table, bind)
+    table.drop(bind=bind, checkfirst=True)
+    if table.exists(bind=bind):
         raise ValueError('table %r dropped but still exists' % table.name)
 
 
 @convert.register(pd.DataFrame, (sa.sql.Select, sa.sql.Selectable), cost=200.0)
-def select_or_selectable_to_frame(el, **kwargs):
-    columns, rows = batch(el)
+def select_or_selectable_to_frame(el, bind=None, **kwargs):
+    columns, rows = batch(el, bind=bind)
     row = next(rows, None)
     if row is None:
         return pd.DataFrame(columns=columns)
@@ -608,82 +607,125 @@ def select_or_selectable_to_frame(el, **kwargs):
 
 class CopyToCSV(sa.sql.expression.Executable, sa.sql.ClauseElement):
 
-    def __init__(self, element, path, delimiter=',', quotechar='"',
-                 lineterminator=r'\n', escapechar='\\', header=True,
-                 na_value=''):
+    def __init__(
+        self,
+        element,
+        path,
+        delimiter=',',
+        quotechar='"',
+        lineterminator='\n',
+        escapechar='\\',
+        header=True,
+        na_value='',
+        encoding=None,
+        bind=None,
+    ):
         self.element = element
         self.path = path
         self.delimiter = delimiter
         self.quotechar = quotechar
         self.lineterminator = lineterminator
+        self._bind = bind = getbind(element, bind)
 
         # mysql cannot write headers
-        self.header = header and element.bind.dialect.name != 'mysql'
+        self.header = header and bind.dialect.name != 'mysql'
         self.escapechar = escapechar
         self.na_value = na_value
+        self.encoding = encoding
 
     @property
     def bind(self):
-        return self.element.bind
+        return self._bind
 
 
 @compiles(CopyToCSV, 'postgresql')
 def compile_copy_to_csv_postgres(element, compiler, **kwargs):
     selectable = element.element
-    istable = isinstance(selectable, sa.Table)
-    template = """COPY %s TO '{path}'
-        WITH CSV {header}
-        DELIMITER '{delimiter}'
-        QUOTE '{quotechar}'
-        NULL '{na_value}'
-        ESCAPE '{escapechar}'
-    """ % ('{query}' if istable else '({query})')
-    processed = (fullname(selectable, compiler)
-                 if istable else compiler.process(selectable))
-    assert processed, ('got empty string from processing element of type %r' %
-                       type(selectable).__name__)
-    return template.format(query=processed,
-                           path=element.path,
-                           header='HEADER' if element.header else '',
-                           delimiter=element.delimiter,
-                           quotechar=element.quotechar,
-                           na_value=element.na_value,
-                           escapechar=element.escapechar)
+    return compiler.process(
+        sa.text(
+            """COPY {0} TO :path
+            WITH (
+                FORMAT CSV,
+                HEADER :header,
+                DELIMITER :delimiter,
+                QUOTE :quotechar,
+                NULL :na_value,
+                ESCAPE :escapechar,
+                ENCODING :encoding
+            )
+            """.format(
+                compiler.preparer.format_table(selectable)
+                if isinstance(selectable, sa.Table)
+                else '({0})'.format(compiler.process(selectable))
+            )
+        ).bindparams(
+            path=element.path,
+            header=element.header,
+            delimiter=element.delimiter,
+            quotechar=element.quotechar,
+            na_value=element.na_value,
+            escapechar=element.escapechar,
+            encoding=element.encoding or element.bind.execute(
+                'show client_encoding'
+            ).scalar()
+        ),
+        **kwargs
+    )
 
 
 @compiles(CopyToCSV, 'mysql')
 def compile_copy_to_csv_mysql(element, compiler, **kwargs):
     selectable = element.element
-    if isinstance(selectable, sa.Table):
-        processed = 'SELECT * FROM %(table)s' % dict(table=selectable.name)
-    else:
-        processed = compiler.process(selectable)
-    assert processed, ('got empty string from processing element of type %r' %
-                       type(selectable).__name__)
-    template = """{query} INTO OUTFILE '{path}'
-    FIELDS TERMINATED BY '{delimiter}'
-    OPTIONALLY ENCLOSED BY '{quotechar}'
-    ESCAPED BY '{escapechar}'
-    LINES TERMINATED BY '{lineterminator}'"""
-    return template.format(query=processed,
-                           path=element.path,
-                           delimiter=element.delimiter,
-                           lineterminator=element.lineterminator,
-                           escapechar=element.escapechar.encode(
-                               'unicode-escape').decode(),
-                           quotechar=element.quotechar)
+    return compiler.process(
+        sa.text(
+            """{0} INTO OUTFILE :path
+            CHARACTER SET :encoding
+            FIELDS TERMINATED BY :delimiter
+            OPTIONALLY ENCLOSED BY :quotechar
+            ESCAPED BY :escapechar
+            LINES TERMINATED BY :lineterminator
+            """.format(
+                compiler.process(
+                    selectable.select()
+                    if isinstance(selectable, sa.Table) else selectable,
+                    **kwargs
+                )
+            )
+        ).bindparams(
+            path=element.path,
+            encoding=element.encoding or element.bind.execute(
+                'select @@character_set_client'
+            ).scalar(),
+            delimiter=element.delimiter,
+            quotechar=element.quotechar,
+            escapechar=element.escapechar,
+            lineterminator=element.lineterminator
+        )
+    )
 
 
 @compiles(CopyToCSV, 'sqlite')
 def compile_copy_to_csv_sqlite(element, compiler, **kwargs):
+    if element.encoding is not None:
+        raise ValueError(
+            "'encoding' keyword argument not supported for "
+            "SQLite to CSV conversion"
+        )
     if not find_executable('sqlite3'):
         raise MDNotImplementedError("Could not find sqlite executable")
 
+    # we are sending a SQL string directorly to the SQLite process so we always
+    # need to bind everything before sending it
+    kwargs['literal_binds'] = True
+
     selectable = element.element
-    sql = (compiler.process(sa.select([selectable])
-                            if isinstance(selectable, sa.Table)
-                            else selectable) + ';')
-    sql = re.sub(r'\s{2,}', ' ', re.sub(r'\s*\n\s*', ' ', sql)).encode()
+    sql = compiler.process(
+        selectable.select() if isinstance(selectable, sa.Table) else selectable,
+        **kwargs
+    ) + ';'
+    sql = re.sub(r'\s{2,}', ' ', re.sub(r'\s*\n\s*', ' ', sql)).encode(
+        sys.getfilesystemencoding()  # we send bytes to the process
+    )
     cmd = ['sqlite3', '-csv',
            '-%sheader' % ('no' if not element.header else ''),
            '-separator', element.delimiter,
@@ -696,11 +738,16 @@ def compile_copy_to_csv_sqlite(element, compiler, **kwargs):
 
 
 @append.register(CSV, sa.sql.Selectable)
-def append_table_to_csv(csv, selectable, dshape=None, **kwargs):
+def append_table_to_csv(csv, selectable, dshape=None, bind=None, **kwargs):
     kwargs = keyfilter(keywords(CopyToCSV).__contains__,
                        merge(csv.dialect, kwargs))
-    stmt = CopyToCSV(selectable, os.path.abspath(csv.path), **kwargs)
-    with selectable.bind.begin() as conn:
+    stmt = CopyToCSV(
+        selectable,
+        os.path.abspath(csv.path),
+        bind=bind,
+        **kwargs
+    )
+    with getbind(selectable, bind).begin() as conn:
         conn.execute(stmt)
     return csv
 
