@@ -7,6 +7,7 @@ pytest.importorskip('psycopg2')
 
 import os
 import itertools
+import shutil
 
 from datashape import dshape
 
@@ -16,22 +17,35 @@ from odo.utils import assert_allclose, tmpfile
 
 
 names = ('tbl%d' % i for i in itertools.count())
+new_schema = object()
 data = [(1, 2), (10, 20), (100, 200)]
 null_data = [(1, None), (10, 20), (100, 200)]
 
 
-@pytest.yield_fixture(scope='module')
+@pytest.yield_fixture
 def csv():
+    s = '\n'.join(','.join(map(str, row)) for row in data).encode('utf8')
     with tmpfile('.csv') as fn:
-        with open(fn, 'w') as f:
-            f.write('\n'.join(','.join(map(str, row)) for row in data))
+        with open(fn, 'wb') as f:
+            f.write(s)
         yield CSV(fn)
 
 
-@pytest.fixture
+@pytest.yield_fixture
+def encoding_csv():
+    path = os.path.join(os.path.dirname(__file__), 'encoding.csv')
+    with tmpfile('.csv') as fn:
+        with open(fn, 'wb') as f, open(path, 'r') as g:
+            f.write(g.read().encode('latin1'))
+        yield CSV(fn)
+
+
+@pytest.yield_fixture
 def complex_csv():
-    this_dir = os.path.dirname(__file__)
-    return CSV(os.path.join(this_dir, 'dummydata.csv'), has_header=True)
+    path = os.path.join(os.path.dirname(__file__), 'dummydata.csv')
+    with tmpfile('.csv') as fn:
+        shutil.copy(path, fn)
+        yield CSV(fn, has_header=True)
 
 
 @pytest.fixture
@@ -39,139 +53,144 @@ def url():
     return 'postgresql://postgres@localhost/test::%s' % next(names)
 
 
-@pytest.yield_fixture
-def sql(url):
-    try:
-        t = resource(url, dshape='var * {a: int32, b: ?int32}')
-    except sa.exc.OperationalError as e:
-        pytest.skip(str(e))
-    else:
+def sql_fixture(dshape, schema=None):
+    @pytest.yield_fixture(params=(True, False))
+    def fixture(request, url):
         try:
-            yield t
-        finally:
-            drop(t)
+            t = resource(
+                url,
+                dshape=dshape,
+                schema=next(names) if schema is new_schema else schema,
+            )
+        except sa.exc.OperationalError as e:
+            pytest.skip(str(e))
+        else:
+            try:
+                bind = t.bind
+                if not request.param:
+                    t.metadata.bind = None
+                yield t, bind
+            finally:
+                t.metadata.bind = bind
+                drop(t)
+                if t.schema:
+                    bind.execute(sa.sql.ddl.DropSchema(t.schema))
+
+    return fixture
 
 
-@pytest.yield_fixture
-def sql_with_schema(url):
-    try:
-        t = resource(url, dshape='var * {a: int32, b: ?int32}',
-                     schema=next(names))
-    except sa.exc.OperationalError as e:
-        pytest.skip(str(e))
-    else:
-        try:
-            yield t
-        finally:
-            drop(t)
-
-
-@pytest.yield_fixture
-def sql_with_ugly_schema(url):
-    try:
-        t = resource(url, dshape='var * {a: int32, b: ?int32}',
-                     schema='foo.b.ar')
-    except sa.exc.OperationalError as e:
-        pytest.skip(str(e))
-    else:
-        try:
-            yield t
-        finally:
-            drop(t)
-            t.bind.execute(sa.sql.ddl.DropSchema(t.schema))
-
-
-@pytest.yield_fixture
-def complex_sql(url):
-    ds = """var * {
+sql = sql_fixture('var * {a: int32, b: ?int32}')
+sql_with_schema = sql_fixture('var * {a: int32, b: ?int32}', new_schema)
+sql_with_ugly_schema = sql_fixture('var * {a: int32, b: ?int32}', 'foo.b.ar')
+complex_sql = sql_fixture("""
+    var * {
         Name: string, RegistrationDate: date, ZipCode: int32, Consts: float64
-    }"""
+    }
+""")
+
+
+@pytest.yield_fixture
+def quoted_sql(csv):
+    url = 'postgresql://postgres@localhost/test::foo bar'
     try:
-        t = resource(url, dshape=ds)
+        t = resource(url, dshape=discover(csv))
     except sa.exc.OperationalError as e:
         pytest.skip(str(e))
     else:
-        yield t
-        drop(t)
+        try:
+            yield t
+        finally:
+            drop(t)
 
 
 def test_simple_into(csv, sql):
-    into(sql, csv, dshape=discover(sql))
-    assert into(list, sql) == data
+    sql, bind = sql
+    into(sql, csv, dshape=discover(sql), bind=bind)
+    assert into(list, sql, bind=bind) == data
 
 
 def test_append(csv, sql):
-    into(sql, csv)
-    assert into(list, sql) == data
+    sql, bind = sql
+    into(sql, csv, bind=bind)
+    assert into(list, sql, bind=bind) == data
 
-    into(sql, csv)
-    assert into(list, sql) == data + data
+    into(sql, csv, bind=bind)
+    assert into(list, sql, bind=bind) == data + data
 
 
 def test_tryexcept_into(csv, sql):
+    sql, bind = sql
     with pytest.raises(sa.exc.NotSupportedError):
-        into(sql, csv, quotechar="alpha")  # uses multi-byte character
+        # uses multi-byte character
+        into(sql, csv, quotechar="alpha", bind=bind)
 
 
 def test_no_header_no_columns(csv, sql):
-    into(sql, csv, dshape=discover(sql))
-    assert into(list, sql) == data
+    sql, bind = sql
+    into(sql, csv, bind=bind, dshape=discover(sql))
+    assert into(list, sql, bind=bind) == data
 
 
 def test_complex_into(complex_csv, complex_sql):
+    complex_sql, bind = complex_sql
     # data from: http://dummydata.me/generate
-    into(complex_sql, complex_csv, dshape=discover(complex_sql))
-    assert_allclose(into(list, complex_sql), into(list, complex_csv))
+    into(complex_sql, complex_csv, dshape=discover(complex_sql), bind=bind)
+    assert_allclose(
+        into(list, complex_sql, bind=bind), into(list, complex_csv)
+    )
 
 
 def test_sql_to_csv(sql, csv):
-    sql = odo(csv, sql)
+    sql, bind = sql
+    sql = odo(csv, sql, bind=bind)
     with tmpfile('.csv') as fn:
-        csv = odo(sql, fn)
+        csv = odo(sql, fn, bind=bind)
         assert odo(csv, list) == data
         assert discover(csv).measure.names == discover(sql).measure.names
 
 
 def test_sql_select_to_csv(sql, csv):
-    sql = odo(csv, sql)
+    sql, bind = sql
+    sql = odo(csv, sql, bind=bind)
     query = sa.select([sql.c.a])
     with tmpfile('.csv') as fn:
-        csv = odo(query, fn)
+        csv = odo(query, fn, bind=bind)
         assert odo(csv, list) == [(x,) for x, _ in data]
 
 
 def test_invalid_escapechar(sql, csv):
+    sql, bind = sql
     with pytest.raises(ValueError):
-        odo(csv, sql, escapechar='12')
+        odo(csv, sql, escapechar='12', bind=bind)
 
     with pytest.raises(ValueError):
-        odo(csv, sql, escapechar='')
+        odo(csv, sql, escapechar='', bind=bind)
 
 
 def test_csv_output_is_not_quoted_by_default(sql, csv):
-    sql = odo(csv, sql)
+    sql, bind = sql
+    sql = odo(csv, sql, bind=bind)
     expected = "a,b\n1,2\n10,20\n100,200\n"
     with tmpfile('.csv') as fn:
-        csv = odo(sql, fn)
+        csv = odo(sql, fn, bind=bind)
         with open(fn, 'rt') as f:
             result = f.read()
         assert result == expected
 
 
 def test_na_value(sql, csv):
-    sql = odo(null_data, sql)
+    sql, bind = sql
+    sql = odo(null_data, sql, bind=bind)
     with tmpfile('.csv') as fn:
-        csv = odo(sql, fn, na_value='NA')
+        csv = odo(sql, fn, na_value='NA', bind=bind)
         with open(csv.path, 'rt') as f:
             raw = f.read()
     assert raw == 'a,b\n1,NA\n10,20\n100,200\n'
 
 
-def test_different_encoding(url):
-    encoding = 'latin1'
-    path = os.path.join(os.path.dirname(__file__), 'encoding.csv')
+def test_different_encoding(url, encoding_csv):
     try:
-        sql = odo(path, url, encoding=encoding)
+        sql = odo(encoding_csv, url, encoding='latin1')
     except sa.exc.OperationalError as e:
         pytest.skip(str(e))
     else:
@@ -192,14 +211,27 @@ def test_different_encoding(url):
 
 
 def test_schema(csv, sql_with_schema):
-    assert odo(odo(csv, sql_with_schema), list) == data
+    sql_with_schema, bind = sql_with_schema
+    assert odo(odo(csv, sql_with_schema, bind=bind), list, bind=bind) == data
 
 
 def test_ugly_schema(csv, sql_with_ugly_schema):
-    assert odo(odo(csv, sql_with_ugly_schema), list) == data
+    sql_with_ugly_schema, bind = sql_with_ugly_schema
+    assert (
+        odo(odo(csv, sql_with_ugly_schema, bind=bind), list, bind=bind) ==
+        data
+    )
 
 
 def test_schema_discover(sql_with_schema):
+    sql_with_schema, bind = sql_with_schema
+    sql_with_schema.metadata.bind = bind
     meta = discover(sql_with_schema.metadata)
     assert meta == dshape('{%s: var * {a: int32, b: ?int32}}' %
                           sql_with_schema.name)
+
+
+def test_quoted_name(quoted_sql, csv):
+    s = odo(csv, quoted_sql)
+    t = odo(csv, list)
+    assert sorted(odo(s, list)) == sorted(t)
